#include "SDL_pad.h"
#include "SDL_padinternal.h"

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

static int pad_state[PADKEY_NUM];
static int* pad_user_state;

Pad_UserButton* Pad_user_buttons;
int Pad_user_button_num;

Pad_UserAxis* Pad_user_axises;
int Pad_user_axis_num;

enum { JOYSTICK_NUM = 2 };
static SDL_Joystick* joysticks[JOYSTICK_NUM];

static void Pad_UserButtonsInit(int to) {
	int i, j;
	pad_user_state = (int*)realloc(pad_user_state, sizeof(int)*to);
	Pad_user_buttons =
		(Pad_UserButton*)realloc(Pad_user_buttons, sizeof(Pad_UserButton)*to);
	for (i = Pad_user_button_num; i < to; i++) {
		pad_user_state[i] = 0;
		for (j = 0; j < USER_BUTTON_BUTTON_NUM; j++) {
			Pad_user_buttons[i][j] = PAD_UNUSED;
		}
	}
	Pad_user_button_num = to;
}

static void Pad_UserAxisesInit(int to) {
	int i, j;
	Pad_user_axises =
		(Pad_UserAxis*)realloc(Pad_user_axises, sizeof(Pad_UserAxis)*to);
	for (i = Pad_user_axis_num; i < to; i++) {
		for (j = 0; j < 4; j++) {
			Pad_user_axises[i][j] = PAD_UNUSED;
		}
	}
	Pad_user_axis_num = to;
}

int Pad_Init() {
	int i;
	for (i = 0; i < PADKEY_NUM; i++) {
		pad_state[i] = 0;
	}

	Pad_user_button_num = 0;
	pad_user_state = NULL;
	Pad_user_buttons = NULL;
	Pad_UserButtonsInit(10);

	Pad_user_axises = NULL;
	Pad_user_axis_num = 0;
	Pad_UserAxisesInit(2);

	Pad_InitStore();

	if (SDL_WasInit(SDL_INIT_EVERYTHING) == 0) {
		Pad_SetError("SDL_Pad: SDL not inited yet");
		return -1;
	}

	if (SDL_WasInit(SDL_INIT_JOYSTICK)) {
		for (i = 0; i < JOYSTICK_NUM; i++) joysticks[i] = NULL;
		for (i = 0; i < SDL_NumJoysticks(); i++) {
			joysticks[i] = SDL_JoystickOpen(i);
		}
	}

    for (i = 0; i < PAD_JOYSTICK_NUM; i++) {
        Pad_SetJoystickThresholdUp(i, 8192);
        Pad_SetJoystickThresholdRight(i, 8192);
        Pad_SetJoystickThresholdDown(i, 8192);
        Pad_SetJoystickThresholdLeft(i, 8192);
    }

    return 0;
}

#define PAD_PRESS_RELEASE(push, state) \
if (push) { \
if (state <= 0) (state) = 1; \
else ++(state); \
} \
else if ((state) <= 0) --(state); \
else (state) = 0;

void Pad_Pump(Uint32 flags) {
	SDL_PumpEvents();

	if (flags | PAD_PUMP_KEYBOARD) {
		Uint8* k = SDL_GetKeyState(NULL);
		int* p = pad_state;
		Uint8* kend = k + SDLK_LAST;
		for (; k < kend; k++) {
			PAD_PRESS_RELEASE(*k, *p);
			p++;
		}
	}

	if (flags | PAD_PUMP_JOYSTICK) {
		int i;
		for (i = 0; i < JOYSTICK_NUM; i++) {
			if (!joysticks[i]) continue;

			{
				Sint16 x, y;
				int base = PADJ1_AXIS - i * 20;
				x = Pad_JoystickGetAxis(joysticks[i], 0, 0);
				y = Pad_JoystickGetAxis(joysticks[i], 1, 0);
				PAD_PRESS_RELEASE((x > 0), pad_state[-base - PAD_AXIS_RIGHT]);
				PAD_PRESS_RELEASE((x < 0), pad_state[-base - PAD_AXIS_LEFT]);
				PAD_PRESS_RELEASE((y > 0), pad_state[-base - PAD_AXIS_DOWN]);
				PAD_PRESS_RELEASE((y < 0), pad_state[-base - PAD_AXIS_UP]);
			}

			{
				int j;
				int base = PADJ1_BUTTON - i * 20;
				for (j = 0; j < SDL_JoystickNumButtons(joysticks[i]); j++) {
					int b = Pad_JoystickGetButton(joysticks[i], j);
					PAD_PRESS_RELEASE(b, pad_state[-base-j]);
				}
			}
		}
	}

	if (flags | PAD_PUMP_MOUSE) {
		Uint8 state = SDL_GetMouseState(NULL, NULL);
		int i;
		for (i = 1; i < 5; i++) {
			PAD_PRESS_RELEASE(SDL_BUTTON(i)&state, pad_state[370+i]);
		}
	}

	if (flags | PAD_PUMP_EVENT) {
		/* this number is from SDL_events.c */
		enum { MAXEVENTS = 128 };
		static SDL_Event e[MAXEVENTS];
		int num, i;
		int quit_found = 0;

		num = SDL_PeepEvents(e, MAXEVENTS, SDL_PEEKEVENT, SDL_QUITMASK);

		for (i = 0; i < num; i++) {
			switch (e[i].type) {
			case (SDL_QUIT):
				quit_found = 1;
				break;
			default:
				Pad_SetError("SDL_Pad: Broken event filter?");
				break;
			}
		}

		PAD_PRESS_RELEASE(quit_found, pad_state[-PADE_QUIT]);
	}

	if (flags | PAD_PUMP_USER) {
		int i, j, p;
		for (i = 0; i < Pad_user_button_num; i++) {
			p = 0;
			for (j = 0; j < USER_BUTTON_BUTTON_NUM; j++) {
				if (Pad_user_buttons[i][j] == PAD_UNUSED) break;
				if (pad_state[-Pad_user_buttons[i][j]] > 0) {
					p = 1;
					break;
				}
			}
			PAD_PRESS_RELEASE(p, pad_user_state[i]);
		}
	}
}

int Pad_GetButton(int id) {
	if (id < 0) return pad_state[-id];
	else return pad_user_state[id];
}

int Pad_GetButtonPushed(int id) {
	return Pad_GetButton(id) == 1;
}
int Pad_GetButtonPressed(int id) {
	return Pad_GetButton(id) > 0;
}
int Pad_GetButtonPulled(int id) {
	return Pad_GetButton(id) == 0;
}
int Pad_GetButtonReleased(int id) {
	return Pad_GetButton(id) <= 0;
}

int Pad_AddButton(int id, int addid) {
	int i;

	if (addid >= 0) {
		Pad_SetError("SDL_Pad: illegal button addid %d", addid);
		return -1;
	}

	if (id >= Pad_user_button_num) {
		int next = Pad_user_button_num * 2;
		while  (id >= next) next *= 2;
		Pad_UserButtonsInit(next);
	}

	for (i = 0; i < USER_BUTTON_BUTTON_NUM; i++) {
		if (Pad_user_buttons[id][i] == PAD_UNUSED) break;
	}
	if (i == USER_BUTTON_BUTTON_NUM) {
		Pad_SetError("SDL_Pad: too many buttons in user button %d\n", id);
		return -1;
	}

	Pad_user_buttons[id][i] = addid;

	return 0;
}

int Pad_GetAxis(int id) {
	return
		((Pad_GetButton(Pad_user_axises[id][0]) > 0) ? PAD_AXIS_UP : 0) |
		((Pad_GetButton(Pad_user_axises[id][1]) > 0) ? PAD_AXIS_DOWN : 0) |
		((Pad_GetButton(Pad_user_axises[id][2]) > 0) ? PAD_AXIS_LEFT : 0) |
		((Pad_GetButton(Pad_user_axises[id][3]) > 0) ? PAD_AXIS_RIGHT : 0);
}

void Pad_AddAxis(int id, int upid, int rightid, int downid, int leftid) {
	if (id >= Pad_user_axis_num) {
		int next = Pad_user_axis_num * 2;
		while  (id >= next) next *= 2;
		Pad_UserAxisesInit(next);
	}

	Pad_user_axises[id][0] = upid;
	Pad_user_axises[id][1] = downid;
	Pad_user_axises[id][2] = leftid;
	Pad_user_axises[id][3] = rightid;
}

void Pad_AddAxisKeyboardArrow(int id) {
	Pad_AddAxis(id, PADK_UP, PADK_RIGHT, PADK_DOWN, PADK_LEFT);
}
void Pad_AddAxisKeyboardRogue(int id) {
	Pad_AddAxis(id, PADK_k, PADK_l, PADK_j, PADK_h);
}
void Pad_AddAxisKeyboardWASD(int id) {
	Pad_AddAxis(id, PADK_w, PADK_d, PADK_s, PADK_a);
}
void Pad_AddAxisJoystick1(int id) {
	Pad_AddAxis(id, PADJ1_UP, PADJ1_RIGHT, PADJ1_DOWN, PADJ1_LEFT);
}
void Pad_AddAxisJoystick2(int id) {
	Pad_AddAxis(id, PADJ2_UP, PADJ2_RIGHT, PADJ2_DOWN, PADJ2_LEFT);
}



// ̾Ƥ
PadKey Pad_PadKeyFromEvent(SDL_Event* ev) {
    switch (ev->type) {
    case SDL_KEYDOWN:
        return - ev->key.keysym.sym;
    case SDL_QUIT:
        return PADE_QUIT;
    case SDL_JOYAXISMOTION: {
        int joy = ev->jaxis.which;
        int id = PADJ1_AXIS - joy * 20;
        int fid = id;
        if (ev->jaxis.axis == 0) {
            if (ev->jaxis.value < -Pad_thre_minus[joy][0])
                id += PAD_AXIS_LEFT;
            else if (ev->jaxis.value > Pad_thre_plus[joy][0])
                id += PAD_AXIS_RIGHT;
        }
        else if (ev->jaxis.axis == 1) {
            if (ev->jaxis.value < -Pad_thre_minus[joy][1])
                id += PAD_AXIS_UP;
            else if (ev->jaxis.value > Pad_thre_plus[joy][1])
                id += PAD_AXIS_DOWN;
        }
        if (id == fid) return PAD_UNUSED;
        else return id;
    }
    case SDL_JOYBUTTONDOWN:
        return PADJ1_BUTTON - ev->jbutton.which * 20 + ev->jbutton.button;
    case SDL_MOUSEBUTTONDOWN:
        return PADM_BUTTON - ev->button.button;
    default:
        return PAD_UNUSED;
    }
}

const int Pad_4axises[] = {
    PAD_AXIS_UP, PAD_AXIS_RIGHT, PAD_AXIS_DOWN, PAD_AXIS_LEFT
};
const int Pad_8axises[] = {
    PAD_AXIS_UP, PAD_AXIS_UPRIGHT, PAD_AXIS_RIGHT, PAD_AXIS_DOWNRIGHT,
    PAD_AXIS_DOWN, PAD_AXIS_DOWNLEFT, PAD_AXIS_LEFT, PAD_AXIS_UPLEFT
};
const char* Pad_axis_names[] = {
    "NONE", "UP", "DOWN", "ERR", "LEFT", "UPLEFT", "DOWNLEFT", "ERR",
    "RIGHT", "UPRIGHT", "DOWNRIGHT", "ERR", "ERR", "ERR", "ERR", "ERR"
};

