#include "SDL_pad.h"
#include "SDL_padinternal.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

char* Pad_default_bnames[] = {
	"A_BUTTON", "B_BUTTON", "C_BUTTON", "D_BUTTON", "E_BUTTON",
	"UP_BUTTON", "RIGHT_BUTTON", "DOWN_BUTTON", "LEFT_BUTTON",
	"END_BUTTON", NULL
};

char* Pad_default_anames[] = {
	"AXIS1", "AXIS2", NULL
};

char* Pad_keynames[-PAD_LAST];

static int find_id(const char* key) {
	int i;
	for (i = 1; i < -PAD_LAST; i++) {
		if (Pad_keynames[i] && strcmp(Pad_keynames[i], key) == 0) {
			return -i;
		}
	}
	return PAD_UNUSED;
}

void Pad_InitStore() {
	int i;
	for (i = 0; i < -PAD_LAST; i++) {
		Pad_keynames[i] = NULL;
	}
#include "SDL_padkeynames.tab"
}

int Pad_LoadDefault(const char* file) {
	return Pad_Load(file, Pad_default_bnames, Pad_default_anames);
}

int Pad_Load(const char* file, char** bnames, char** anames) {
	int ret;
	SDL_RWops* op = SDL_RWFromFile(file, "r");

	if (!op) return -1;

	ret = Pad_LoadRW(op, bnames, anames);
	SDL_RWclose(op);

	return ret;
}

int Pad_LoadRW(SDL_RWops* op, char** bnames, char** anames) {
	static const int BUFSIZE = 256;
	char buf[USER_BUTTON_BUTTON_NUM+1][BUFSIZE];
	int btnite = 0;
	int bufite = 0;
	char c;
	int i, j;

	while (SDL_RWread(op, &c, 1, 1)) {
		if (c == '\n' || c == '#') {
			if (bufite != 0) {
				buf[btnite][bufite] = '\0';
				btnite++;
			}

			if (btnite > 1) {
                if (strstr(buf[0], "PADCONFIG_") == buf[0]) {
                    for (i = 0; i < PAD_JOYSTICK_NUM; i++) {
                        for (j = 0; j < 4; j++) {
                            char buf2[256];
                            int a = Pad_4axises[j];
                            sprintf(buf2, "JOYSTICK%d_THRESHOULD_%s",
                                    i, Pad_axis_names[a]);
                            if (strstr(buf[0]+10, buf2) == buf[0]+10) {
                                Pad_SetJoystickThreshold(i, atoi(buf[1]), a);
                                goto endadd;
                            }
                        }
                    }
                    Pad_SetError("SDL_Pad: invalid PADCONFIG: %s", buf[0]);
                    return -1;
                }

				char** n = anames;
				int id = PAD_UNUSED;
				i = 0;
				for (; *n; n++) {
					if (strcmp(buf[0], *n) == 0) {
						id = i;
						break;
					}
					i++;
				}

                /* axises */
				if (id != PAD_UNUSED) {
					int axisid[4];

					if (btnite != 5) goto endadd;

					for (i = 1; i < 5; i++) {
						int addid = PAD_UNUSED;
						int j = 0;
						char** n = bnames;

						for (; *n; n++) {
							if (strcmp(buf[i], *n) == 0) {
								addid = j;
								break;
							}
							j++;
						}

						if (addid == PAD_UNUSED) {
							addid = find_id(buf[i]);
						}

						if (addid == PAD_UNUSED) goto endadd;

						axisid[i-1] = addid;
					}

					Pad_AddAxis(id, axisid[0],axisid[1],axisid[2],axisid[3]);
				}
                /* user buttons */
				else {
					char** n = bnames;
					int id = PAD_UNUSED;
					i = 0;
					for (; *n; n++) {
						if (strcmp(buf[0], *n) == 0) {
							id = i;
							break;
						}
						i++;
					}

					if (id != PAD_UNUSED) {
						for (i = 1; i < btnite; i++) {
							int addid = find_id(buf[i]);
							if (addid != 0) {
								Pad_AddButton(id, addid);
							}
						}
					}
				}
			}

		endadd:
			btnite = 0;
			bufite = 0;

			if (c == '#') {
				while (SDL_RWread(op, &c, 1, 1)) {
					if (c == '\n') break;
				}
			}
		}
		else if (isalnum(c) || c == '_') {
			buf[btnite][bufite] = c;
			bufite++;
			if (bufite >= BUFSIZE) {
				Pad_SetError("SDL_Pad: too long line read");
				return -1;
			}
		}
		/* the other is treated as delimiter */
		else {
			if (bufite) {
				buf[btnite][bufite] = '\0';
				btnite++;
				bufite = 0;
			}
		}

	}

	return 0;
}

int Pad_SaveDefault(const char* file) {
	return Pad_Save(file, Pad_default_bnames, Pad_default_anames);
}

int Pad_Save(const char* file, char** bnames, char** anames) {
	int ret;
	SDL_RWops* op = SDL_RWFromFile(file, "w");

	if (!op) return -1;

	ret = Pad_SaveRW(op, bnames, anames);
	SDL_RWclose(op);

	return ret;
}

int Pad_SaveRW(SDL_RWops* op, char** bnames, char** anames) {
	static const int BUFSIZE = 1024;
	char buf[BUFSIZE];
	int i = 0, j;
	int name_size;
	char** n = bnames;

	for (; *n && i < Pad_user_button_num; n++, i++) {
		if (Pad_user_buttons[i][0] == PAD_UNUSED) continue;

		strncpy(buf, *n, BUFSIZE);
		strncat(buf, " =", BUFSIZE-strlen(buf));

		for (j = 0; j < USER_BUTTON_BUTTON_NUM; j++) {
			int b = Pad_user_buttons[i][j];
			char* name = NULL;
			if (b == PAD_UNUSED) break;
			name = Pad_keynames[-b];
			if (name) {
				strncat(buf, " ", BUFSIZE-strlen(buf));
				strncat(buf, name, BUFSIZE-strlen(buf));
			}
		}

		strncat(buf, "\n", BUFSIZE-strlen(buf));

		SDL_RWwrite(op, buf, 1, strlen(buf));
	}

	name_size = i;

	for (i = 0; i < Pad_user_axis_num && i < 9; i++) {
		if (anames[i] == NULL) break;

		for (j = 0; j < 4; j++) {
			if (Pad_user_axises[i][j] == PAD_UNUSED) break;
		}
		if (j != 4) continue;

		strncpy(buf, anames[i], BUFSIZE);
		strncat(buf, " =", BUFSIZE-strlen(buf));

		for (j = 0; j < 4; j++) {
			int id = Pad_user_axises[i][j];
			char* name = NULL;

			if (id < 0) {
				name = Pad_keynames[-id];
			}
			else if (id < name_size) {
				name = bnames[id];
			}
			else {
				/* @todo it's error */
				break;
			}

			/* @todo it's error */
			if (!name) break;

			strncat(buf, " ", BUFSIZE-strlen(buf));
			strncat(buf, name, BUFSIZE-strlen(buf));
		}

		strncat(buf, "\n", BUFSIZE-strlen(buf));

		SDL_RWwrite(op, buf, 1, strlen(buf));
	}

    for (i = 0; i < PAD_JOYSTICK_NUM; i++) {
        sprintf(buf, "PADCONFIG_JOYSTICK%d_THRESHOULD_RIGHT=%d\n",
                i, Pad_thre_plus[i][0]);
        SDL_RWwrite(op, buf, 1, strlen(buf));
        sprintf(buf, "PADCONFIG_JOYSTICK%d_THRESHOULD_LEFT=%d\n",
                i, Pad_thre_minus[i][0]);
        SDL_RWwrite(op, buf, 1, strlen(buf));
        sprintf(buf, "PADCONFIG_JOYSTICK%d_THRESHOULD_UP=%d\n",
                i, Pad_thre_minus[i][1]);
        SDL_RWwrite(op, buf, 1, strlen(buf));
        sprintf(buf, "PADCONFIG_JOYSTICK%d_THRESHOULD_DOWN=%d\n",
                i, Pad_thre_plus[i][1]);
        SDL_RWwrite(op, buf, 1, strlen(buf));
    }

	return 0;
}


